use crossterm::event::KeyCode;
use readmap::types::Section;
use std::path::PathBuf;
use tool_helper::Error;
use tui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Color, Modifier, Style},
    text::{Span, Spans},
    widgets::{Block, Borders, BorderType, Paragraph, Tabs}
};

pub type EventReceiver = std::sync::mpsc::Receiver<Event<crossterm::event::KeyEvent>>;
pub type Terminal      = tui::Terminal<tui::backend::CrosstermBackend<std::io::Stdout>>;

pub enum Event<I> {
    Input(I),
    Tick,
    ForceRender,
}

pub enum UIState {
    Alive,
    Render,
    Terminated
}

pub enum MenuSelection {
    Stats,
    Quit,
}

impl MenuSelection {
    fn next(&self) -> MenuSelection {
        match self {
            Self::Stats => Self::Quit,
            Self::Quit  => Self::Stats
        }
    }

    fn prev(&self) -> MenuSelection {
        self.next()
    }

    fn get_id(&self) -> usize {
        match self {
            MenuSelection::Stats => 0,
            MenuSelection::Quit  => 1,
        }
    }
}

pub struct ConsoleUI {
    rx:             EventReceiver,
    terminal:       Terminal,
    frame:          Rect,
    sub_frames:     Vec<Rect>,
    menu_selection: MenuSelection,
}

impl ConsoleUI {
    pub fn new(rx: EventReceiver, terminal: Terminal) -> ConsoleUI {
        ConsoleUI{rx, terminal, frame: Rect::default(), sub_frames: Vec::default(), menu_selection: MenuSelection::Stats}
    }

    pub fn update(&mut self) -> Result<UIState, Error> {
        match self.rx.recv()? {
            Event::Input(event) => {
                match event.code {
                    KeyCode::Char('q') => Ok(UIState::Terminated),
                    KeyCode::Left      => {
                        self.menu_selection = self.menu_selection.prev();
                        Ok(UIState::Render)
                    },
                    KeyCode::Right     => {
                        self.menu_selection = self.menu_selection.next();
                        Ok(UIState::Render)
                    },
                    KeyCode::Enter     => {
                        if matches!(self.menu_selection, MenuSelection::Quit) {
                            Ok(UIState::Terminated)
                        }
                        else {
                            Ok(UIState::Alive)
                        }
                    },
                    _                  => Ok(UIState::Alive)
                }
            },
            Event::Tick        => Ok(UIState::Alive),
            Event::ForceRender => Ok(UIState::Render)
        }
    }

    pub fn render(&mut self) -> Result<(), Error> {
        self.terminal.draw(|frame| {
            Self::update_sub_frames(&mut self.sub_frames, &mut self.frame, frame.size());

            frame.render_widget(Self::create_titel(),                    self.sub_frames[0]);
            frame.render_widget(Self::create_menu(&self.menu_selection), self.sub_frames[1]);
            frame.render_widget({
                match self.menu_selection {
                    MenuSelection::Stats => Self::create_titel(),
                    MenuSelection::Quit  => Self::create_exit_message(),
                }
            }, self.sub_frames[2]);
        })?;
        Ok(())
    }

    pub fn close(mut self) -> Result<(), Error> {
        crossterm::terminal::disable_raw_mode()?;
        self.terminal.show_cursor()?;
        self.terminal.clear()?;

        Ok(())
    }

    fn update_sub_frames(sub_frames: &mut Vec<Rect>, frame: &mut Rect, new_frame: Rect) {
        if new_frame  != *frame {
            *sub_frames = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Length(3),
                Constraint::Length(3),
                Constraint::Min(3)
            ]).split(new_frame);
            *frame = new_frame;
        }
    }

    fn create_titel<'a>() -> Paragraph<'a> {
        Paragraph::new("psxreadmap")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain)
            )
    }

    fn create_menu<'a>(menu_selection: &MenuSelection) -> Tabs<'a> {
        const MENU_TITLES: &'static [&'static str] = &["Stats", "Quit"];

        let menu = MENU_TITLES.iter().map(|t| {
            let (first, rest) = t.split_at(1);
            Spans::from(vec![
                Span::styled(first, Style::default().fg(Color::Yellow).add_modifier(Modifier::UNDERLINED),),
                Span::styled(rest,  Style::default().fg(Color::White)),
            ])
        }).collect();
        
        Tabs::new(menu).select(menu_selection.get_id()).block(
            Block::default()
            .title("Menu").borders(Borders::ALL))
            .style(Style::default().fg(Color::White))
            .highlight_style(Style::default().bg(Color::LightYellow))
            .divider(Span::raw("|")
        )
    }

    fn create_exit_message<'a>() -> Paragraph<'a> {
        Paragraph::new("Press \"ENTER\" to exit")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain).title("QUIT")
            )
    }
}

pub fn load_memory_map(file_path: Option<PathBuf>) -> Result<Vec<Section>, Error> {
    readmap::scan(tool_helper::open_input(file_path)?)
}